import io
import os
import pandas as pd

def put_file(session, stage, file, filename=None):
    # Create an in-memory file object
    file_obj = io.BytesIO(file)

    # Create a temporary file in the filesystem and upload it
    import tempfile
    import shutil

    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        shutil.copyfileobj(file_obj, temp_file)
        temp_file_path = temp_file.name

    # Use Snowflake's PUT command to upload the temporary file to the stage
    if filename:
        put_command = f"PUT file://{temp_file_path} @{stage}/{filename}"
    else:
        put_command = f"PUT file://{temp_file_path} @{stage}"

    # Use the correct Snowflake session method
    try:
        # Use Snowpark's sql().collect() method
        session.sql(put_command).collect()
    except Exception as e:
        raise Exception(f"Failed to execute PUT command: {put_command}. Error: {e}")

    # Clean up temporary file
    os.remove(temp_file_path)

    # Note: Removed session.close() to allow reuse of session for multiple operations

def put_dataframe_to_stage(session, stage_name, df, filename=None, schema=None):
    """
    Write a DataFrame directly to a Snowflake stage without using temporary files.
    This is designed for notebook environments where ephemeral storage is not reliable.
    
    Args:
        session: Snowflake Snowpark session
        stage_name: Name of the Snowflake stage (without @ symbol)
        df: pandas DataFrame to write
        filename: Optional filename for the stage file
        schema: Optional schema name for temporary table creation
    
    Returns:
        str: The filename that was written to the stage
    """
    if filename is None:
        import uuid
        filename = f"data_{uuid.uuid4().hex[:8]}.csv"
    
    # Convert DataFrame to CSV string with proper data type handling
    # Convert all columns to string to avoid type conversion issues
    df_string = df.astype(str)
    csv_buffer = io.StringIO()
    df_string.to_csv(csv_buffer, index=False)
    csv_content = csv_buffer.getvalue()
    
    # Create a Snowflake DataFrame from the CSV content
    # We'll use a temporary table approach to get the data into the stage
    temp_table_name = f"TEMP_{filename.replace('.csv', '').replace('-', '_')}"
    
    # Get current database for fully qualified table name
    current_db = session.sql('SELECT CURRENT_DATABASE()').collect()[0][0]
    
    # Use schema if provided, otherwise use current schema
    if schema:
        full_temp_table_name = f"{current_db}.{schema}.{temp_table_name}"
    else:
        full_temp_table_name = f"{current_db}.{temp_table_name}"
    
    try:
        # Write DataFrame to a temporary table using centralized data_writer
        from . import data_writer
        data_writer.write_dataframe_to_table(
            session=session,
            df=df,
            schema=schema if schema else "PUBLIC",
            table=temp_table_name,
            auto_create=True,
            overwrite=True,
            use_logical_type=True,
            on_error="CONTINUE"
        )
        
        # Copy from temporary table to stage
        copy_command = f"""
        COPY INTO @{stage_name}/{filename}
        FROM {full_temp_table_name}
        FILE_FORMAT = (TYPE = 'CSV' FIELD_DELIMITER = ',' FIELD_OPTIONALLY_ENCLOSED_BY = '"' 
                      NULL_IF = ('NULL', 'null') EMPTY_FIELD_AS_NULL = TRUE)
        HEADER = TRUE
        OVERWRITE = TRUE
        """
        
        session.sql(copy_command).collect()
        
        # Clean up temporary table
        session.sql(f"DROP TABLE IF EXISTS {full_temp_table_name}").collect()
        
        return filename
        
    except Exception as e:
        # Clean up temporary table on error
        try:
            session.sql(f"DROP TABLE IF EXISTS {full_temp_table_name}").collect()
        except:
            pass
        raise Exception(f"Failed to write DataFrame to stage {stage_name}: {e}")

def put_csv_content_to_stage(session, stage_name, csv_content, filename=None, schema=None):
    """
    Write CSV content directly to a Snowflake stage without using temporary files.
    This is designed for notebook environments where ephemeral storage is not reliable.
    
    Args:
        session: Snowflake Snowpark session
        stage_name: Name of the Snowflake stage (without @ symbol)
        csv_content: String containing CSV data
        filename: Optional filename for the stage file
        schema: Optional schema name for temporary table creation
    
    Returns:
        str: The filename that was written to the stage
    """
    if filename is None:
        import uuid
        filename = f"data_{uuid.uuid4().hex[:8]}.csv"
    
    # Convert CSV string to DataFrame with proper data type handling
    csv_buffer = io.StringIO(csv_content)
    df = pd.read_csv(csv_buffer, low_memory=False, dtype=str)
    
    # Use the DataFrame method to write to stage
    return put_dataframe_to_stage(session, stage_name, df, filename, schema)


import logging
from typing import Optional, Dict, Any
from snowflake.snowpark import Session
from snowflake.snowpark.dataframe import DataFrame
import pandas as pd
import numpy as np
from . import table_creator

logger = logging.getLogger(__name__)

def validate_dataframe_types(df: pd.DataFrame, expected_types: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
    """
    Validate DataFrame data types and identify potential casting issues.
    
    Args:
        df: DataFrame to validate
        expected_types: Optional dict of column_name: expected_type pairs
        
    Returns:
        Dict containing validation results and recommendations
    """
    validation_result = {
        'is_valid': True,
        'issues': [],
        'recommendations': [],
        'column_analysis': {}
    }
    
    for column in df.columns:
        col_data = df[column]
        col_analysis = {
            'dtype': str(col_data.dtype),
            'null_count': col_data.isnull().sum(),
            'unique_count': col_data.nunique(),
            'sample_values': col_data.dropna().head(5).tolist(),
            'mixed_types': False,
            'potential_issues': []
        }
        
        # Check for mixed types in the column
        if col_data.dtype == 'object':
            # Check if object column contains mixed types
            non_null_values = col_data.dropna()
            if len(non_null_values) > 0:
                # Check if all values can be converted to the same type
                try:
                    # Try to convert to numeric
                    pd.to_numeric(non_null_values, errors='raise')
                    col_analysis['potential_issues'].append("Object column contains numeric data - consider converting to numeric type")
                except (ValueError, TypeError):
                    # Check if all values are strings
                    if not all(isinstance(x, str) for x in non_null_values):
                        col_analysis['mixed_types'] = True
                        col_analysis['potential_issues'].append("Mixed data types detected - may cause Snowflake casting issues")
        
        # Check for specific problematic patterns
        if col_data.dtype == 'object':
            # Look for numeric strings that might cause casting issues
            numeric_strings = 0
            total_non_null = len(col_data.dropna())
            
            for value in col_data.dropna():
                if isinstance(value, str) and value.replace('.', '').replace('-', '').replace('e', '').replace('E', '').isdigit():
                    numeric_strings += 1
            
            if numeric_strings > 0 and numeric_strings < total_non_null:
                col_analysis['potential_issues'].append(f"Mixed string/numeric values: {numeric_strings}/{total_non_null} are numeric strings")
        
        validation_result['column_analysis'][column] = col_analysis
        
        # Check for critical issues
        if col_analysis['mixed_types'] or len(col_analysis['potential_issues']) > 0:
            validation_result['is_valid'] = False
            validation_result['issues'].append(f"Column '{column}': {', '.join(col_analysis['potential_issues'])}")
    
    # Generate recommendations
    if not validation_result['is_valid']:
        validation_result['recommendations'].extend([
            "Consider using explicit data type conversion before writing to Snowflake",
            "Use the 'use_logical_type=False' option in write_pandas for more lenient type handling",
            "Check for mixed data types in problematic columns and standardize them",
            "Consider using the 'on_error=ABORT' option to fail fast on type issues"
        ])
    
    return validation_result

def write_dataframe_to_table(
    session: Session, 
    df: pd.DataFrame, 
    schema: str, 
    table: str, 
    overwrite: bool = False, 
    auto_create: bool = True,
    temp_table: Optional[str] = None,
    validate_types: bool = True,
    use_logical_type: bool = True,
    on_error: str = "CONTINUE",
    df_fields: Optional[dict] = None,
    snowflake_fields: Optional[dict] = None,
    force_full_sync: bool = False
) -> bool:
    """
    Centralized function for writing DataFrames to Snowflake tables.
    
    Args:
        session: Snowflake Snowpark session
        df: DataFrame to write
        schema: Target schema name
        table: Target table name
        overwrite: Whether to overwrite existing data
        auto_create: Whether to auto-create the table if it doesn't exist
        temp_table: Optional temporary table name for batch operations
        validate_types: Whether to validate DataFrame types before writing
        use_logical_type: Whether to use logical types (set to False for lenient type handling)
        on_error: Error handling strategy ("CONTINUE", "ABORT", "SKIP_FILE")
        df_fields: Optional dictionary of field types for formatting (uses field_types.format_sync_file if provided)
        
    Returns:
        bool: True if successful, False otherwise
        
    Raises:
        Exception: If the write operation fails
    """
    try:
        # Format DataFrame using field_types.format_sync_file if df_fields provided
        if df_fields is not None:
            try:
                from lht.util import field_types
                df = field_types.format_sync_file(df, df_fields)
                
                # VALIDATION: Ensure snowflake_fields is provided when df_fields is available
                if snowflake_fields is None or len(snowflake_fields) == 0:
                    logger.error(f"❌ CRITICAL: df_fields provided but snowflake_fields is missing!")
                    logger.error(f"   This means sobjects.describe() failed or wasn't called")
                    logger.error(f"   Cannot proceed without proper Salesforce field type definitions")
                    raise Exception(f"Missing snowflake_fields when df_fields is provided for {schema}.{table}")
                    
            except Exception as e:
                logger.warning(f"⚠️ Failed to format DataFrame with field_types: {e}")
                logger.warning(f"   Continuing with original DataFrame")
        
        # Validate DataFrame types if requested
        if validate_types:
            validation_result = validate_dataframe_types(df)
            if not validation_result['is_valid']:
                logger.warning("⚠️ DataFrame type validation found potential issues:")
                for issue in validation_result['issues']:
                    logger.warning(f"   - {issue}")
                for recommendation in validation_result['recommendations']:
                    logger.info(f"   💡 {recommendation}")
                
                # If there are critical issues, consider using more lenient settings
                if use_logical_type and any("Mixed data types" in issue for issue in validation_result['issues']):
                    logger.info("🔄 Switching to lenient type handling due to mixed data types")
                    use_logical_type = False
                    on_error = "CONTINUE"  # Be more permissive with errors
        
        # Get current database for fully qualified table name
        current_db = session.sql('SELECT CURRENT_DATABASE()').collect()[0][0]
        
        # Determine the target table name
        if temp_table:
            target_table = temp_table
        else:
            target_table = table
            
        full_table_name = f"{current_db}.{schema}.{target_table}"
        
        logger.debug(f"💾 Writing DataFrame to table: {full_table_name}")
        logger.debug(f"   - Records: {len(df)}")
        logger.debug(f"   - Overwrite: {overwrite}")
        logger.debug(f"   - Auto-create: {auto_create}")
        logger.debug(f"   - Use logical types: {use_logical_type}")
        logger.debug(f"   - On error: {on_error}")
        
        # If we have field definitions and need to create a table, use centralized table creation
        if auto_create and df_fields is not None and len(df_fields) > 0:
            try:
                # The DataFrame has already been processed by field_types.format_sync_file above
                # No need to process it again - use the already-processed DataFrame
                
                # Use provided snowflake_fields if available, otherwise infer from DataFrame
                if snowflake_fields is not None and len(snowflake_fields) > 0:
                    # Use the Salesforce field types directly (correct approach)
                    pass
                else:
                    # No Salesforce field types provided - this should not happen in normal operation
                    logger.error(f"❌ CRITICAL: No Salesforce field types provided for table creation!")
                    logger.error(f"   This means sobjects.describe() failed or wasn't called")
                    logger.error(f"   Cannot create table without proper field type definitions")
                    raise Exception(f"No Salesforce field types provided for table {schema}.{table}")
                

                # Use centralized table creation
                print(f"🔍 About to call table_creator.ensure_table_exists_for_dataframe with force_full_sync={force_full_sync}")
                table_creator.ensure_table_exists_for_dataframe(
                    session=session,
                    schema=schema,
                    table=target_table,
                    df_fields=df_fields,
                    snowflake_fields=snowflake_fields,
                    force_full_sync=force_full_sync,  # Pass through the force_full_sync parameter
                    database=current_db
                )
                print(f"✅ Table creation completed, now writing data with overwrite={overwrite}")
                
                # Now write to the existing table using the already-processed DataFrame
                # Convert all datetime fields to timezone-naive for Snowflake compatibility
                for col in df.columns:
                    if col.upper() in ['CREATEDDATE', 'LASTMODIFIEDDATE', 'SYSTEMMODSTAMP', 'LASTACTIVITYDATE', 'LASTVIEWEDDATE', 'LASTREFERENCEDDATE']:
                        if 'UTC' in str(df[col].dtype) or 'timezone' in str(df[col].dtype):
                            try:
                                df[col] = df[col].dt.tz_localize(None)
                            except Exception as e:
                                logger.error(f"❌ Failed to convert {col} to timezone-naive: {e}")
                
                #df['CREATEDDATE'] = pd.to_datetime(df['CREATEDDATE'], errors='coerce')
                # Convert to Snowpark DataFrame and use save_as_table instead of write_pandas
                print(f"🔍 Writing data with mode: {'overwrite' if overwrite else 'append'}")
                print(f"🔍 Target table: {full_table_name}")
                snowpark_df = session.create_dataframe(df)
                df = None
                snowpark_df.write.mode("overwrite" if overwrite else "append").save_as_table(full_table_name)
                snowpark_df = None
                print(f"✅ Data written successfully with mode: {'overwrite' if overwrite else 'append'}")
                result = True  # save_as_table doesn't return a result object like write_pandas
            except Exception as table_error:
                logger.warning(f"Centralized table creation failed, falling back to auto-create: {table_error}")
                # CRITICAL FIX: Use the processed DataFrame for fallback, not the original
                logger.info(f"🔧 Using processed DataFrame for fallback write...")
                try:
                    from lht.util import field_types
                    df_fallback = field_types.format_sync_file(df, df_fields)
                    logger.info(f"✅ DataFrame processing completed for fallback write")
                except Exception as e:
                    logger.warning(f"⚠️ Failed to process DataFrame with field_types for fallback: {e}")
                    df_fallback = df
                
                # TEST: Keeping all datetime fields for debugging (fallback path)
                logger.info(f"🧪 TEST: Keeping all datetime fields to debug the issue (fallback path)")
                logger.info(f"✅ TEST: All columns preserved in fallback: {list(df_fallback.columns)}")
                
                # Convert all datetime fields to timezone-naive for Snowflake compatibility
                # for col in df_fallback.columns:
                #     if col.upper() in ['CREATEDDATE', 'LASTMODIFIEDDATE', 'SYSTEMMODSTAMP', 'LASTACTIVITYDATE', 'LASTVIEWEDDATE', 'LASTREFERENCEDDATE']:
                #         if 'UTC' in str(df_fallback[col].dtype) or 'timezone' in str(df_fallback[col].dtype):
                #             try:
                #                 df_fallback[col] = df_fallback[col].dt.tz_localize(None)
                #             except Exception as e:
                #                 logger.error(f"❌ Failed to convert {col} to timezone-naive: {e}")
                
                # Convert to Snowpark DataFrame and use save_as_table instead of write_pandas
                print(f"🔍 Writing fallback data with mode: {'overwrite' if overwrite else 'append'}")
                print(f"🔍 Target table: {full_table_name}")
                snowpark_df = session.create_dataframe(df_fallback)
                df_fallback = None
                snowpark_df.write.mode("overwrite" if overwrite else "append").save_as_table(full_table_name)
                snowpark_df = None
                print(f"✅ Fallback data written successfully with mode: {'overwrite' if overwrite else 'append'}")
                result = True  # save_as_table doesn't return a result object like write_pandas
        else:
            # Use original behavior, but ensure DataFrame is properly processed
            if df_fields is not None and len(df_fields) > 0:
                try:
                    from lht.util import field_types
                    df_processed = field_types.format_sync_file(df, df_fields)
                except Exception as e:
                    logger.warning(f"⚠️ Failed to process DataFrame with field_types: {e}")
                    df_processed = df
            else:
                df_processed = df
            
            # Convert all datetime fields to timezone-naive for Snowflake compatibility
            for col in df_processed.columns:
                if col.upper() in ['CREATEDDATE', 'LASTMODIFIEDDATE', 'SYSTEMMODSTAMP', 'LASTACTIVITYDATE', 'LASTVIEWEDDATE', 'LASTREFERENCEDDATE']:
                    if 'UTC' in str(df_processed[col].dtype) or 'timezone' in str(df_processed[col].dtype):
                        try:
                            df_processed[col] = df_processed[col].dt.tz_localize(None)
                        except Exception as e:
                            logger.error(f"❌ Failed to convert {col} to timezone-naive: {e}")
            
            # Convert to Snowpark DataFrame and use save_as_table instead of write_pandas
            print(f"🔍 Writing processed data with mode: {'overwrite' if overwrite else 'append'}")
            print(f"🔍 Target table: {full_table_name}")
            snowpark_df = session.create_dataframe(df_processed)
            df_processed = None
            snowpark_df.write.mode("overwrite" if overwrite else "append").save_as_table(full_table_name)
            snowpark_df = None
            print(f"✅ Processed data written successfully with mode: {'overwrite' if overwrite else 'append'}")
            result = True  # save_as_table doesn't return a result object like write_pandas
        
        logger.debug(f"✅ Successfully wrote {len(df)} records to {full_table_name}")
        return True
        
    except Exception as e:
        logger.error(f"❌ Failed to write DataFrame to {schema}.{table}: {e}")
        logger.error(f"   - Error type: {type(e).__name__}")
        
        # Provide specific guidance for common errors
        if "Failed to cast" in str(e):
            logger.error("   🔍 This appears to be a data type casting issue.")
            logger.error("   💡 Try using validate_types=True and use_logical_type=False")
            logger.error("   💡 Check for mixed data types in your DataFrame columns")
        
        raise Exception(f"DataFrame write operation failed: {e}")




def standardize_dataframe_types(df: pd.DataFrame, type_strategy: str = "auto") -> pd.DataFrame:
    """
    Standardize DataFrame data types to prevent Snowflake casting issues.
    
    Args:
        df: DataFrame to standardize
        type_strategy: Strategy for type conversion ("auto", "string", "numeric", "lenient")
        
    Returns:
        DataFrame with standardized types
    """
    logger.info(f"🔧 Standardizing DataFrame types using strategy: {type_strategy}")
    
    df_standardized = df.copy()
    
    for column in df_standardized.columns:
        col_data = df_standardized[column]
        original_dtype = str(col_data.dtype)
        
        if type_strategy == "string":
            # Convert everything to string
            df_standardized[column] = col_data.astype(str)
            
        elif type_strategy == "numeric":
            # Try to convert to numeric where possible, string otherwise
            try:
                df_standardized[column] = pd.to_numeric(col_data, errors='coerce')
            except (ValueError, TypeError):
                df_standardized[column] = col_data.astype(str)
                
        elif type_strategy == "lenient":
            # More intelligent type conversion
            if col_data.dtype == 'object':
                # Check if column contains mostly numeric data
                non_null_values = col_data.dropna()
                if len(non_null_values) > 0:
                    numeric_count = 0
                    total_count = len(non_null_values)
                    
                    for value in non_null_values:
                        if isinstance(value, str):
                            # Check if string represents a number
                            try:
                                float(value)
                                numeric_count += 1
                            except ValueError:
                                pass
                        elif isinstance(value, (int, float)) or (hasattr(value, 'is_integer') and value.is_integer()):
                            numeric_count += 1
                    
                    # If more than 80% are numeric, convert the whole column
                    if numeric_count / total_count > 0.8:
                        try:
                            df_standardized[column] = pd.to_numeric(col_data, errors='coerce')
                        except (ValueError, TypeError):
                            df_standardized[column] = col_data.astype(str)
                    else:
                        df_standardized[column] = col_data.astype(str)
                        
        else:  # "auto" strategy
            # Smart type inference
            if col_data.dtype == 'object':
                # Try to infer the best type
                non_null_values = col_data.dropna()
                if len(non_null_values) > 0:
                    # Check if all values can be converted to the same type
                    all_numeric = True
                    all_strings = True
                    
                    for value in non_null_values:
                        if isinstance(value, str):
                            try:
                                float(value)
                            except ValueError:
                                all_numeric = False
                        elif not isinstance(value, (int, float)):
                            all_numeric = False
                            all_strings = False
                    
                    if all_numeric:
                        try:
                            df_standardized[column] = pd.to_numeric(col_data, errors='coerce')
                        except (ValueError, TypeError):
                            df_standardized[column] = col_data.astype(str)
                    elif all_strings:
                        df_standardized[column] = col_data.astype(str)
                    else:
                        # Mixed types - convert to string to be safe
                        df_standardized[column] = col_data.astype(str)
    
    logger.info(f"✅ DataFrame type standardization completed")
    return df_standardized

def write_dataframe_with_type_handling(
    session: Session,
    df: pd.DataFrame,
    schema: str,
    table: str,
    type_strategy: str = "auto",
    **kwargs
) -> bool:
    """
    Write DataFrame to Snowflake with automatic type handling and validation.
    
    Args:
        session: Snowflake Snowpark session
        df: DataFrame to write
        schema: Target schema name
        table: Target table name
        type_strategy: Strategy for type standardization ("auto", "string", "numeric", "lenient")
        **kwargs: Additional arguments for write_dataframe_to_table
        
    Returns:
        bool: True if successful
    """
    try:
        # Standardize types first
        df_standardized = standardize_dataframe_types(df, type_strategy)
        
        # Use lenient settings for problematic data
        if type_strategy in ["string", "lenient"]:
            kwargs.setdefault('use_logical_type', False)
            kwargs.setdefault('on_error', 'CONTINUE')
        
        # Write the standardized DataFrame
        return write_dataframe_to_table(
            session=session,
            df=df_standardized,
            schema=schema,
            table=table,
            **kwargs
        )
        
    except Exception as e:
        logger.error(f"❌ Failed to write DataFrame with type handling: {e}")
        raise

def write_batch_to_temp_table(
    session: Session,
    df: pd.DataFrame,
    schema: str,
    temp_table: str,
    df_fields: dict,
    validate_types: bool = True,
    force_full_sync: bool = False
) -> bool:
    """
    Write a batch DataFrame to a temporary table for incremental sync operations.
    
    Args:
        session: Snowflake Snowpark session
        df: DataFrame to write
        schema: Target schema name
        temp_table: Temporary table name
        df_fields: List of field names for validation
        validate_types: Whether to validate DataFrame types before writing
        
    Returns:
        bool: True if successful
    """
    logger.debug(f"📦 Writing batch to temp table: {schema}.{temp_table}")
    
    try:
        return write_dataframe_to_table(
            session=session,
            df=df,
            schema=schema,
            table=temp_table,
            overwrite=False,
            auto_create=False,
            temp_table=temp_table,
            validate_types=validate_types,
            use_logical_type=False,  # More lenient for temp tables
            on_error="CONTINUE",
            df_fields=df_fields,  # Pass field definitions for proper formatting
            force_full_sync=force_full_sync  # Pass through force_full_sync parameter
        )
    except Exception as e:
        error_msg = str(e)
        if any(phrase in error_msg for phrase in ["Failed to cast", "cast", "variant", "FIXED"]):
            logger.warning(f"⚠️ Casting error detected in temp table write: {error_msg[:100]}...")
            logger.warning(f"⚠️ Retrying with type standardization...")
            
                        # Standardize types and retry
            df_standardized = standardize_dataframe_types(df, "string")
            return write_dataframe_to_table(
                session=session,
                df=df_standardized,
                schema=schema,
                table=temp_table,
                overwrite=False,
                auto_create=False,
                temp_table=temp_table,
                validate_types=False,
                use_logical_type=False,
                on_error="CONTINUE",
                df_fields=df_fields,  # Pass field definitions for proper formatting
                force_full_sync=force_full_sync  # Pass through force_full_sync parameter
            )
        else:
            logger.error(f"❌ Non-casting error in temp table write: {error_msg}")
            raise

def write_batch_to_main_table(
    session: Session,
    df: pd.DataFrame,
    schema: str,
    table: str,
    is_first_batch: bool = False,
    validate_types: bool = True,
    use_logical_type: bool = True,
    df_fields: Optional[dict] = None,
    snowflake_fields: Optional[dict] = None,
    force_full_sync: bool = False
) -> bool:
    """
    Write a batch DataFrame to the main table (overwrite for first batch, append for subsequent).
    
    Args:
        session: Snowflake Snowpark session
        df: DataFrame to write
        schema: Target schema name
        table: Target table name
        is_first_batch: Whether this is the first batch (determines overwrite behavior)
        validate_types: Whether to validate DataFrame types before writing
        use_logical_type: Whether to use logical types (set to False for lenient type handling)
        
    Returns:
        bool: True if successful
    """
    # Handle overwrite logic based on force_full_sync and batch type
    if force_full_sync and is_first_batch:
        # Force full sync: drop and recreate table, then overwrite first batch
        overwrite = True
        print(f"💾 Force full sync: overwriting first batch to recreate table")
    elif is_first_batch:
        # Regular first batch: append to existing or new table
        overwrite = False
        print(f"💾 Regular first batch: appending to table")
    else:
        # Subsequent batches: always append
        overwrite = False
        print(f"💾 Subsequent batch: appending to table")
    
    print(f"💾 Writing batch to main table: {schema}.{table} (overwrite={overwrite}, is_first_batch={is_first_batch}, force_full_sync={force_full_sync})")
    
    return write_dataframe_to_table(
        session=session,
        df=df,
        schema=schema,
        table=table,
        overwrite=overwrite,
        auto_create=True,
        validate_types=validate_types,
        use_logical_type=use_logical_type,
        on_error="CONTINUE",
        df_fields=df_fields,  # Pass field definitions for proper formatting
        snowflake_fields=snowflake_fields,  # Pass Salesforce field types for proper table creation
        force_full_sync=force_full_sync  # Pass through force_full_sync parameter
    )



import json
import requests
from snowflake.snowpark import Session
from snowflake.snowpark.functions import col
from . import csv


def job(session, json_data):   
    if 'externalIdFieldName' in json_data:
        externalid = json_data['externalIdFieldName']
    else:
        externalid = None
    query = """INSERT INTO LOGS.RETL_HISTORY (id, operation, object, createdById, createdDate, externalIdFieldName,contentUrl) values(\'{}\',\'{}\',\'{}\',\'{}\',\'{}\',\'{}\',\'{}\')""".format(json_data['id'], json_data['operation'], json_data['object'], json_data['createdById'], json_data['createdDate'][:10]+' '+json_data['createdDate'][11:23], externalid,json_data['contentUrl'])
    session.sql(query).collect()
    return None

def successful_results(access_info, job_id):
    access_token = access_info['access_token']
    url = access_info['instance_url']+f"/services/data/v62.0/jobs/ingest/{job_id}/successfulResults/"
    headers = {
        'Authorization': f'Bearer {access_token}',
        'Accept': 'text/csv'
    }
    response = requests.get(url, headers=headers)
    results = csv.success_upserts(response.text, job_id)

    return results

def failed_results(access_info, job_id):
    access_token = access_info['access_token']
    url = access_info['instance_url']+f"/services/data/v62.0/jobs/ingest/{job_id}/failedResults/"
    headers = {
        'Authorization': f'Bearer {access_token}',
        'Accept': 'text/csv'
    }
    response = requests.get(url, headers=headers)
    results = csv.fail_upserts(response.text, job_id)

    return results

def log_results(session, access_info, job_id, schema):   

    successes = successful_results(access_info, job_id)
    if len(successes) > 0:
        session.write_pandas(successes, 'RETL_RESULTS', schema=schema, quote_identifiers=False, auto_create_table=False, overwrite=False,use_logical_type=True)

    failures = failed_results(access_info, job_id)
    if len(failures) > 0:
        session.write_pandas(failures, 'RETL_FAILURES', schema=schema, quote_identifiers=False, auto_create_table=False, overwrite=False,use_logical_type=True)

    return None


def check_log(session, id):
    query = history_query(id)
    results = session.sql(query).collect() 
    
    return len(results)

def history_query(id):
    query = """with history as (
                select 
                ID, 
                OPERATION, 
                OBJECT, 
                EXTERNALIDFIELDNAME
                from RADNET_SANDBOX.LOGS.RETL_HISTORY
                ),
                results as (
                Select 
                HISTORY_ID
                from RADNET_SANDBOX.LOGS.RETL_RESULTS
                )
                select
                h.id,
                r.history_id
                from history h
                join results r
                on r.history_id = h.id
                where h.id = \'{}\'""".format(id)

    return query


from lht.salesforce import sobjects as sobj

def build_soql(session, access_info, sobject):
    query_string, df_fields, create_table_fields = sobj.describe(session, access_info, sobject)
    query_string = "select"
    first_time = True
    for field in df_fields.keys():
        if first_time == True:
            query_string = query_string + "\n"+field
            first_time = False
        else:
            query_string = query_string + '\n,'+field
    query_string = query_string + "\n from {}".format(sobject)
    return query_string


from snowflake.snowpark import Session
from snowflake.snowpark import functions as F

def format_filter_condition(snowpark_session, src_table, tgt_table,src_filter, tgt_filter):
  filter_cond = list()
  split_src = src_filter.split(',')
  split_tgt = tgt_filter.split(',')
  
  # -- Check Both source and targer filter condition are same length
  # -- Note : Filter condition order matters here:
  if len(split_src) == len(split_tgt):
    for i in range(len(split_src)):
        #print(i)
        filter_cond.append('src.'+ '"' + split_src[i]+ '"' + '= tgt.'+ '"' +split_tgt[i]+ '"')
  else:
    return "Error"
  
  s_filter_cond = " AND ".join(filter_cond)
  
  # -- Call the function to generate the merge statement
  s_merge_stament = format_insert_upsert(snowpark_session, src_table, tgt_table, s_filter_cond)
  
  # -- Execute the Merge Statement
  s_final_result = ""
  #if s_merge_stament.upper() != 'ERROR':
  if s_merge_stament != 'Error':
    #print("\n\n@@@ {}".format(s_merge_stament))
    src_table_col = snowpark_session.sql(s_merge_stament).collect()
    s_final_result = "Number of Rows Inserted: {0} Updated:{1}".format(str(src_table_col[0][0]), str(src_table_col[0][1]))
    
  else:
    print("error")
    s_final_result = "Error"
  
  return s_final_result;
 
def format_insert_upsert(snowpark_session, src_table, tgt_table, s_filter_cond):
    """
        Function query the snowflake metadata and generate the Merge
    """
    sel_colum = list()
    update_col = list()
    insert_sel = list()
    insert_val = list()
    # Get current database and schema for proper table references
    current_db = snowpark_session.sql("SELECT CURRENT_DATABASE()").collect()[0][0]
    schema_name = snowpark_session.get_current_schema().replace('"','')
    
    # Use fully qualified table names for temporary table queries
    src_table_col = snowpark_session.sql("select COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = '{0}' AND TABLE_SCHEMA = \'{1}\' ORDER BY ORDINAL_POSITION".format(src_table, schema_name)).collect()
    tgt_table_col = snowpark_session.sql("select COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = '{0}' AND TABLE_SCHEMA = \'{1}\' ORDER BY ORDINAL_POSITION".format(tgt_table, schema_name)).collect()

    if len(src_table_col) != 0:
        for idx_value in range(len(src_table_col)):
            sel_colum.append('"'+src_table_col[idx_value][0]+'"')
            insert_val.append("src." + '"' + str(src_table_col[idx_value][0]) + '"')
            insert_sel.append("tgt." + '"' + str(tgt_table_col[idx_value][0]) + '"')
            update_col.append("tgt." + '"' + str(tgt_table_col[idx_value][0]) + '"' + ' = ' + "src." + '"' + str(src_table_col[idx_value][0]) + '"')

        s_merge_stmt = """
                    MERGE INTO
                       {0} tgt
                    USING
                        (SELECT {1} FROM {2}) src
                    ON
                        {3}
                    WHEN MATCHED THEN UPDATE SET
                        {4}
                    WHEN NOT MATCHED THEN INSERT
                         ({5})
                    VALUES
                        ({6})
                """.format(
                    tgt_table,
                    ",".join(sel_colum),
                    src_table,
                    s_filter_cond,
                    ",".join(update_col),
                    ",".join(insert_sel),
                    ",".join(insert_val)
                )   
    else:
        return "Error"
    return s_merge_stmt

#method to transform temp table and match datatypes with permanent table
def transform_and_match_datatypes(session, temp_table, permanent_table, temp_schema=None, perm_schema=None):
  """
  Transform data types in temp table to match permanent table schema.

  Args:
      session: Snowpark session
      temp_table: Name of temporary table
      permanent_table: Name of permanent table  
      temp_schema: Schema of temp table (optional)
      perm_schema: Schema of permanent table (optional)

  Returns:
      Snowpark DataFrame with transformed data types
  """


  fields = ""
  # Get schema info for both tables
  temp_schema_info = session.sql(f"DESCRIBE TABLE {temp_table}").collect()

  perm_schema_info = session.sql(f"DESCRIBE TABLE {permanent_table}").collect()

  # Create mapping of column name to data type for permanent table
  perm_types = {row['name'].upper(): row['type'] for row in perm_schema_info}
  temp_types = {row['name'].upper(): row['type'] for row in temp_schema_info}

  for row in perm_schema_info:
      col_name = row['name'].upper()
      
      if col_name in temp_types:
          temp_type = temp_types[col_name]
          perm_type = perm_types[col_name]
      
          # If types don't match, add cast
          if temp_type != perm_type:
              # Handle common type conversions
              if 'VARCHAR' in perm_type or 'STRING' in perm_type:
                  expr = f"CASE WHEN TRIM({col_name}) = 'nan' THEN NULL ELSE CAST({col_name} AS {perm_type}) END as {col_name},\n"
              elif 'NUMBER' in perm_type or 'DECIMAL' in perm_type or 'FLOAT' in perm_type:
                  expr = f"CASE WHEN {col_name} ='nan' THEN NULL when trim({col_name}) = '' then NULL ELSE {col_name}::NUMBER END as {col_name},\n" 

              elif 'INTEGER' in perm_type or 'BIGINT' in perm_type:
                  expr = f"CASE WHEN {col_name} ='nan'  then NULL when {col_name} = '' then NULL ELSE {col_name}::NUMBER END as {col_name},\n" 
              elif 'TIMESTAMP' in perm_type or 'TIMESTAMP_NTZ' in perm_type:
                #expr = f"CASE WHEN {col_name} ='nan' then NULL when {col_name} = '' then NULL ELSE TO_TIMESTAMP_NTZ(REPLACE({col_name}, 'Z', ''), 'YYYY-MM-DD\"T\"HH24:MI:SS.FF3') END as {col_name},\n"
                expr = f"""CASE 
                    WHEN {col_name} = 'nan' THEN NULL 
                    WHEN {col_name} = '' THEN NULL 
                    ELSE 
                        TO_TIMESTAMP_NTZ(left({col_name},19), 'YYYY-MM-DD\"T\"HH24:MI:SS')
                END as {col_name},\n"""
              elif 'DATE' in perm_type:
                  expr = f"CASE WHEN {col_name} ='nan' then NULL when {col_name} = '' then NULL ELSE TO_DATE(SUBSTR({col_name}, 1, 10), 'YYYY-MM-DD') END as {col_name},\n"
          
              elif 'BOOLEAN' in perm_type:
                  expr = f"{col_name}::boolean AS {col_name},\n"
          
              fields = fields + expr
          else:
              fields = fields + f"CASE WHEN TRIM({col_name}) = 'nan' then NULL when {col_name} = '' THEN NULL ELSE CAST({col_name} AS {perm_type}) END as {col_name},\n"
      else:
          fields = fields + 'Null as ' + col_name+",\n"
  return fields


import requests
import json

def job_close(access_info, job_id):
    access_token = access_info['access_token']
    print("closing job")
    url = access_info['instance_url']+f"/services/data/v62.0/jobs/ingest/{job_id}/"
    headers = {
        'Authorization': f'Bearer {access_token}',
        'Content-Type': 'application/json'
    }
    close = {"state":"UploadComplete"}
    response = requests.patch(url, headers=headers, data=json.dumps(close))
    #response.raise_for_status()
    print(response.status_code)

def job_status(access_info, job_id):
    access_token = access_info['access_token']
    url = access_info['instance_url']+f"/services/data/v62.0/jobs/ingest/{job_id}"
    headers = {
        'Authorization': f'Bearer {access_token}'
    }
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    print("@@@ STATUS {}".format(response.json()))
    return response.json()

def send_file(access_info, job_id, data):
    access_token = access_info['access_token']
    url = access_info['instance_url']+f"/services/data/v62.0/jobs/ingest/{job_id}/batches/"
    print(url)
    headers = {
        'Authorization': f'Bearer {access_token}',
        'Content-Type': 'text/csv'
    }
    response = requests.put(url, headers=headers, data=data)
    print(response.status_code)




import pandas as pd
from . import sobjects as sobj, sobject_query as sobj_query
from lht.util import merge, data_writer as dw

def new_changed_records(session, access_info, sobject, local_table, match_field, lmd=None):

    if lmd is None:
        #get the most recent last modified date
        local_query = """Select max(LastModifiedDate::timestamp_ntz) as LastModifiedDate from {}""".format(local_table)
        results_df = session.sql(local_query).collect()

        lmd = pd.to_datetime(results_df[0]['LASTMODIFIEDDATE'])

    if lmd is not None:
        lmd_sf = str(pd.to_datetime(lmd))[:10]+'T'+str(pd.to_datetime(lmd))[11:19]+'.000z'
    else:
        lmd_sf = None
    tmp_table = 'TMP_{}'.format(local_table)
    session.sql("CREATE or REPLACE TEMPORARY TABLE {} LIKE {}""".format(tmp_table,local_table)).collect()

    #get the columns from the local table.  There may be fields that are not in the local table
    #and the salesforce sync will need to skip them
    results = session.sql("SHOW COLUMNS IN TABLE TMP_{}".format(local_table)).collect()
    table_fields = []
    for field in results:
        try:
            table_fields.append(field[2]) 
        except:
            print("field {} not present.  Skipping".format(field[2]))
            continue
 
    #method returns the salesforce sobject query and the fields from the sobject
    query, df_fields, snowflake_fields = sobj.describe(access_info, sobject, lmd_sf)

    sobject_data = sobj_query.query_records(access_info, query)
    data_list = list(sobject_data)  # Convert generator to list
    print(f"wadnet_dashboard_data.stg.dashboard_stage/archival-radnet-encounter {len(data_list)} batches")
    for i, batch_df in enumerate(data_list):
        df_str = batch_df.astype(str)
        batch_df = None
        print(f"📊 Processing first batch of data")
        session.write_pandas(df_str, table_name="tmp_"+local_table, auto_create_table=True, overwrite=True, table_type="temporary", quote_identifiers=False)
        df_str = None
        transformed_data = merge.transform_and_match_datatypes(session, "tmp_"+local_table, local_table)
        
        session.sql(f"Insert into {local_table} select {transformed_data} from tmp_{local_table}").collect()

    #merge.format_filter_condition(session, tmp_table, local_table,match_field, match_field)
    return query


import requests
import json
import csv as csv_module
import io
from snowflake.snowpark import Session
from snowflake.snowpark.exceptions import SnowparkSQLException

def create_logs_schema_if_not_exists(session: Session):
    """Create the logs schema if it doesn't exist."""
    try:
        session.sql("CREATE SCHEMA IF NOT EXISTS LOGS").collect()
        print("✅ LOGS schema created/verified successfully")
    except Exception as e:
        print(f"⚠️ Warning: Could not create LOGS schema: {e}")

def create_job_info_table(session: Session):
    """Create the job_info table based on Salesforce API documentation."""
    create_table_sql = """
    CREATE TABLE IF NOT EXISTS LOGS.JOB_INFO (
        ID VARCHAR(255) PRIMARY KEY,
        OPERATION VARCHAR(50),
        OBJECT VARCHAR(255),
        CREATED_BY_ID VARCHAR(255),
        CREATED_DATE TIMESTAMP_NTZ,
        SYSTEM_MODSTAMP TIMESTAMP_NTZ,
        STATE VARCHAR(20),
        JOB_TYPE VARCHAR(50),
        NUMBER_RECORDS_PROCESSED INTEGER,
        NUMBER_RECORDS_FAILED INTEGER,
        RETRY_COUNT INTEGER,
        TOTAL_PROCESSING_TIME INTEGER,
        API_ACTIVE_PROCESSING_TIME INTEGER,
        AVERAGE_PER_SECOND INTEGER,
        AVERAGE_PER_BATCH INTEGER,
        NUMBER_RECORDS_TOTAL INTEGER,
        NUMBER_BATCHES_TOTAL INTEGER,
        NUMBER_BATCHES_IN_PROGRESS INTEGER,
        NUMBER_BATCHES_COMPLETED INTEGER,
        NUMBER_BATCHES_FAILED INTEGER,
        NUMBER_BATCHES_PENDING INTEGER,
        BYTES_PROCESSED INTEGER,
        ERROR_CODE VARCHAR(255),
        ERROR_MESSAGE VARCHAR(16777216),
        CREATED_AT TIMESTAMP_LTZ DEFAULT CURRENT_TIMESTAMP()
    )
    """
    try:
        session.sql(create_table_sql).collect()
        print("✅ JOB_INFO table created/verified successfully")
    except Exception as e:
        print(f"⚠️ Warning: Could not create JOB_INFO table: {e}")

def create_success_table(session: Session):
    """Create the success table based on Salesforce API documentation."""
    create_table_sql = """
    CREATE TABLE IF NOT EXISTS LOGS.SUCCESS (
        ID INTEGER IDENTITY PRIMARY KEY,
        JOB_ID VARCHAR(255) NOT NULL,
        SF_ID VARCHAR(255),
        CREATED TIMESTAMP_NTZ,
        SUCCESS BOOLEAN,
        ERRORS VARCHAR(16777216),
        CREATED_AT TIMESTAMP_LTZ DEFAULT CURRENT_TIMESTAMP()
    )
    """
    try:
        session.sql(create_table_sql).collect()
        print("✅ SUCCESS table created/verified successfully")
    except Exception as e:
        print(f"⚠️ Warning: Could not create SUCCESS table: {e}")

def create_failure_table(session: Session):
    """Create the failure table based on Salesforce API documentation."""
    create_table_sql = """
    CREATE TABLE IF NOT EXISTS LOGS.FAILURE (
        ID INTEGER IDENTITY PRIMARY KEY,
        JOB_ID VARCHAR(255) NOT NULL,
        SF_ID VARCHAR(255),
        SF_ERROR VARCHAR(16777216),
        CREATED_AT TIMESTAMP_LTZ DEFAULT CURRENT_TIMESTAMP()
    )
    """
    try:
        session.sql(create_table_sql).collect()
        print("✅ FAILURE table created/verified successfully")
    except Exception as e:
        print(f"⚠️ Warning: Could not create FAILURE table: {e}")

def get_job_info(access_info: dict, job_id: str) -> dict:
    """Retrieve job information from Salesforce Bulk API 2.0."""
    access_token = access_info['access_token']
    url = f"{access_info['instance_url']}/services/data/v62.0/jobs/ingest/{job_id}"
    
    headers = {
        'Authorization': f'Bearer {access_token}',
        'Accept': 'application/json'
    }
    
    try:
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as e:
        error_info = {
            'error': True,
            'error_code': getattr(e.response, 'status_code', 'UNKNOWN') if hasattr(e, 'response') else 'UNKNOWN',
            'error_message': str(e),
            'state': 'Failed'
        }
        print(f"❌ Error retrieving job info: {error_info}")
        return error_info

def get_successful_results(access_info: dict, job_id: str) -> list:
    """Retrieve successful results from Salesforce Bulk API 2.0."""
    access_token = access_info['access_token']
    url = f"{access_info['instance_url']}/services/data/v62.0/jobs/ingest/{job_id}/successfulResults/"
    
    headers = {
        'Authorization': f'Bearer {access_token}',
        'Accept': 'text/csv'
    }
    
    try:
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        
        # Parse CSV response
        csv_content = response.text
        if not csv_content.strip():
            print("ℹ️ No CSV content received for successful results")
            return []
            
        # Parse CSV and convert to list of dictionaries
        try:
            csv_reader = csv_module.DictReader(io.StringIO(csv_content))
            results = []
            for row in csv_reader:
                results.append(row)
            
            print(f"✅ Successfully parsed {len(results)} successful records from CSV")
            return results
            
        except Exception as csv_error:
            print(f"⚠️ Warning: CSV parsing error for successful results: {csv_error}")
            print(f"📄 CSV content preview: {csv_content[:200]}...")
            return []
        
    except requests.exceptions.RequestException as e:
        print(f"❌ Error retrieving successful results: {e}")
        return []
    except Exception as e:
        print(f"❌ Error parsing successful results CSV: {e}")
        return []

def get_failed_results(access_info: dict, job_id: str) -> list:
    """Retrieve failed results from Salesforce Bulk API 2.0."""
    access_token = access_info['access_token']
    url = f"{access_info['instance_url']}/services/data/v62.0/jobs/ingest/{job_id}/failedResults/"
    
    headers = {
        'Authorization': f'Bearer {access_token}',
        'Accept': 'text/csv'
    }
    
    try:
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        
        # Parse CSV response
        csv_content = response.text
        if not csv_content.strip():
            print("ℹ️ No CSV content received for failed results")
            return []
            
        # Parse CSV and convert to list of dictionaries
        try:
            csv_reader = csv_module.DictReader(io.StringIO(csv_content))
            results = []
            for row in csv_reader:
                results.append(row)
            
            print(f"✅ Successfully parsed {len(results)} failed records from CSV")
            
            # 🔍 DEBUG: Show sample of failed results structure
            if results:
                		print(f"📋 Sample failed result structure:")
                print(f"   Keys: {list(results[0].keys())}")
                print(f"   Sample record: {results[0]}")
                print(f"   Total records: {len(results)}")
            
            return results
            
        except Exception as csv_error:
            print(f"⚠️ Warning: CSV parsing error for failed results: {csv_error}")
            print(f"📄 CSV content preview: {csv_content[:200]}...")
            return []
        
    except requests.exceptions.RequestException as e:
        print(f"❌ Error retrieving failed results: {e}")
        return []
    except Exception as e:
        print(f"❌ Error parsing failed results CSV: {e}")
        return []

def insert_job_info(session: Session, job_data: dict, job_id: str):
    """Insert job information into the JOB_INFO table."""
    try:
        # Handle error cases
        if job_data.get('error'):
            insert_sql = f"""
            INSERT INTO LOGS.JOB_INFO (
                ID, OPERATION, OBJECT, CREATED_BY_ID, CREATED_DATE, SYSTEM_MODSTAMP,
                STATE, JOB_TYPE, NUMBER_RECORDS_PROCESSED, NUMBER_RECORDS_FAILED,
                RETRY_COUNT, TOTAL_PROCESSING_TIME, API_ACTIVE_PROCESSING_TIME,
                AVERAGE_PER_SECOND, AVERAGE_PER_BATCH, NUMBER_RECORDS_TOTAL,
                NUMBER_BATCHES_TOTAL, NUMBER_BATCHES_IN_PROGRESS, NUMBER_BATCHES_COMPLETED,
                NUMBER_BATCHES_FAILED, NUMBER_BATCHES_PENDING, BYTES_PROCESSED,
                ERROR_CODE, ERROR_MESSAGE
            ) VALUES (
                '{job_id}',
                NULL, NULL, NULL, NULL, NULL,
                '{job_data.get('state', 'Failed')}',
                NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                NULL, NULL, NULL, NULL, NULL, NULL,
                '{str(job_data.get('error_code', '')).replace(chr(39), chr(39)+chr(39))}',
                '{str(job_data.get('error_message', '')).replace(chr(39), chr(39)+chr(39))}'
            )
            """
        else:
            # Normal job data - safely escape string values and handle NULLs properly
            operation = f"'{str(job_data.get('operation', '')).replace(chr(39), chr(39)+chr(39))}'" if job_data.get('operation') else 'NULL'
            object_name = f"'{str(job_data.get('object', '')).replace(chr(39), chr(39)+chr(39))}'" if job_data.get('object') else 'NULL'
            created_by_id = f"'{str(job_data.get('createdById', '')).replace(chr(39), chr(39)+chr(39))}'" if job_data.get('createdById') else 'NULL'
            # Format timestamps: replace 'T' with space and remove '+0000'
            created_date_raw = job_data.get('createdDate')
            created_date = f"TO_TIMESTAMP_NTZ(TO_TIMESTAMP_TZ('{created_date_raw.replace('T', ' ').replace('+0000', '')}'))" if created_date_raw else 'NULL'
            
            system_modstamp_raw = job_data.get('systemModstamp')
            system_modstamp = f"TO_TIMESTAMP_NTZ(TO_TIMESTAMP_TZ('{system_modstamp_raw.replace('T', ' ').replace('+0000', '')}'))" if system_modstamp_raw else 'NULL'
            state = f"'{str(job_data.get('state', '')).replace(chr(39), chr(39)+chr(39))}'" if job_data.get('state') else 'NULL'
            job_type = f"'{str(job_data.get('jobType', '')).replace(chr(39), chr(39)+chr(39))}'" if job_data.get('jobType') else 'NULL'
            
            # Handle error fields properly
            error_code = f"'{str(job_data.get('errorCode', '')).replace(chr(39), chr(39)+chr(39))}'" if job_data.get('errorCode') else 'NULL'
            error_message = f"'{str(job_data.get('errorMessage', '')).replace(chr(39), chr(39)+chr(39))}'" if job_data.get('errorMessage') else 'NULL'
            
            insert_sql = f"""
            INSERT INTO LOGS.JOB_INFO (
                ID, OPERATION, OBJECT, CREATED_BY_ID, CREATED_DATE, SYSTEM_MODSTAMP,
                STATE, JOB_TYPE, NUMBER_RECORDS_PROCESSED, NUMBER_RECORDS_FAILED,
                RETRY_COUNT, TOTAL_PROCESSING_TIME, API_ACTIVE_PROCESSING_TIME,
                AVERAGE_PER_SECOND, AVERAGE_PER_BATCH, NUMBER_RECORDS_TOTAL,
                NUMBER_BATCHES_TOTAL, NUMBER_BATCHES_IN_PROGRESS, NUMBER_BATCHES_COMPLETED,
                NUMBER_BATCHES_FAILED, NUMBER_BATCHES_PENDING, BYTES_PROCESSED,
                ERROR_CODE, ERROR_MESSAGE
            ) VALUES (
                '{job_id}',
                {operation},
                {object_name},
                {created_by_id},
                {created_date},
                {system_modstamp},
                {state},
                {job_type},
                {job_data.get('numberRecordsProcessed') or 'NULL'},
                {job_data.get('numberRecordsFailed') or 'NULL'},
                {job_data.get('retryCount') or 'NULL'},
                {job_data.get('totalProcessingTime') or 'NULL'},
                {job_data.get('apiActiveProcessingTime') or 'NULL'},
                {job_data.get('averagePerSecond') or 'NULL'},
                {job_data.get('averagePerBatch') or 'NULL'},
                {job_data.get('numberRecordsTotal') or 'NULL'},
                {job_data.get('numberBatchesTotal') or 'NULL'},
                {job_data.get('numberBatchesInProgress') or 'NULL'},
                {job_data.get('numberBatchesCompleted') or 'NULL'},
                {job_data.get('numberBatchesFailed') or 'NULL'},
                {job_data.get('numberBatchesPending') or 'NULL'},
                {job_data.get('bytesProcessed') or 'NULL'},
                {error_code},
                {error_message}
            )
            """

        
        session.sql(insert_sql).collect()
        print(f"✅ Job info inserted for job {job_id}")
        
    except Exception as e:
        print(f"❌ Error inserting job info: {e}")
        		print(f"📋 SQL that failed: {insert_sql}")

def insert_success_records(session: Session, job_id: str, success_data: list):
    """Insert successful records into the SUCCESS table."""
    if not success_data:
        print("ℹ️ No successful records to insert")
        return
    
    try:
        for record in success_data:
            # Safely escape string values and handle NULLs properly
            sf_id = f"'{str(record.get('sf__Id', '')).replace(chr(39), chr(39)+chr(39))}'" if record.get('sf__Id') else 'NULL'
            
            # Format timestamps: replace 'T' with space and remove '+0000'
            created_raw = record.get('sf__Created')
            created = f"TO_TIMESTAMP_NTZ(TO_TIMESTAMP_TZ('{str(created_raw).replace('T', ' ').replace('+0000', '').replace(chr(39), chr(39)+chr(39))}'))" if created_raw else 'NULL'
            
            success = f"'{str(record.get('sf__Success', '')).replace(chr(39), chr(39)+chr(39))}'" if record.get('sf__Success') else 'NULL'
            errors = f"'{json.dumps(record.get('sf__Errors')).replace(chr(39), chr(39)+chr(39))}'" if record.get('sf__Errors') else 'NULL'
            
            insert_sql = f"""
            INSERT INTO LOGS.SUCCESS (
                JOB_ID, SF_ID, CREATED, SUCCESS, ERRORS
            ) VALUES (
                '{job_id}',
                {sf_id},
                {created},
                {success},
                {errors}
            )
            """
            
            session.sql(insert_sql).collect()
        
        print(f"✅ Inserted {len(success_data)} successful records for job {job_id}")
        
    except Exception as e:
        print(f"❌ Error inserting success records: {e}")

def insert_failure_records(session: Session, job_id: str, failure_data: list):
    """Insert failed records into the FAILURE table."""
    if not failure_data:
        print("ℹ️ No failed records to insert")
        return
    
    		print(f"📋 Starting to insert {len(failure_data)} failed records for job {job_id}")
    
    try:
        for i, record in enumerate(failure_data):
            		print(f"📋 Processing failed record {i+1}/{len(failure_data)}")
            print(f"   Raw record: {record}")
            print(f"   Record keys: {list(record.keys())}")
            
            # Safely escape string values and handle NULLs properly
            sf_id = f"'{str(record.get('sf__Id', '')).replace(chr(39), chr(39)+chr(39))}'" if record.get('sf__Id') else 'NULL'
            
            # Get error information
            sf_error = f"'{str(record.get('sf__Error', '')).replace(chr(39), chr(39)+chr(39))}'" if record.get('sf__Error') else 'NULL'
            
            print(f"   Processed values:")
            print(f"     sf_id: {sf_id}")
            print(f"     sf_error: {sf_error}")
            
            insert_sql = f"""
            INSERT INTO LOGS.FAILURE (
                JOB_ID, SF_ID, SF_ERROR
            ) VALUES (
                '{job_id}',
                {sf_id},
                {sf_error}
            )
            """
            
            print(f"   Generated SQL: {insert_sql}")
            
            try:
                result = session.sql(insert_sql).collect()
                print(f"   ✅ SQL execution successful: {result}")
            except Exception as sql_error:
                print(f"   ❌ SQL execution failed: {sql_error}")
                print(f"   SQL that failed: {insert_sql}")
                raise sql_error
        
        print(f"✅ Inserted {len(failure_data)} failed records for job {job_id}")
        
    except Exception as e:
        print(f"❌ Error inserting failure records: {e}")
        		print(f"📋 Last processed record index: {i if 'i' in locals() else 'N/A'}")
        raise

def process_bulk_api_results(session: Session, access_info: dict, job_id: str):
    """
    Main function to process Bulk API 2.0 job results and store them in the database.
    
    Args:
        session: Snowflake session object
        access_info: Salesforce access credentials dictionary
        job_id: The Bulk API 2.0 job ID to process
    """
    print(f"\n🚀 Processing Bulk API 2.0 results for job: {job_id}")
    print("=" * 80)
    
    try:
        # Step 1: Create schema and tables if they don't exist
        print("📋 Setting up database schema and tables...")
        create_logs_schema_if_not_exists(session)
        create_job_info_table(session)
        create_success_table(session)
        create_failure_table(session)
        
        # Step 2: Retrieve job information
        print(f"📊 Retrieving job information for job {job_id}...")
        job_info = get_job_info(access_info, job_id)
        
        # Step 3: Insert job information
        print("💾 Storing job information...")
        insert_job_info(session, job_info, job_id)
        
        # Step 4: Retrieve and store successful results
        print("📈 Retrieving successful results...")
        success_results = get_successful_results(access_info, job_id)
        insert_success_records(session, job_id, success_results)
        
        # Step 5: Retrieve and store failed results
        print("📉 Retrieving failed results...")
        failure_results = get_failed_results(access_info, job_id)
        
        		print(f"📋 Failure results summary:")
        print(f"   Type: {type(failure_results)}")
        print(f"   Length: {len(failure_results) if failure_results else 0}")
        print(f"   Is empty: {not failure_results}")
        
        if failure_results:
            print(f"   First record keys: {list(failure_results[0].keys())}")
            print(f"   Sample first record: {failure_results[0]}")
        
        insert_failure_records(session, job_id, failure_results)
        
        # Step 6: Summary
        print("\n" + "=" * 80)
        print("✅ BULK API 2.0 RESULTS PROCESSING COMPLETED")
        print(f"📊 Job ID: {job_id}")
        print(f"📈 Successful Records: {len(success_results)}")
        print(f"📉 Failed Records: {len(failure_results)}")
        print(f"🗄️ Data stored in LOGS schema tables")
        print("=" * 80)
        
        return {
            'job_id': job_id,
            'success_count': len(success_results),
            'failure_count': len(failure_results),
            'job_state': job_info.get('state', 'Unknown'),
            'success': True
        }
        
    except Exception as e:
        print(f"\n❌ ERROR processing Bulk API 2.0 results: {e}")
        return {
            'job_id': job_id,
            'success_count': 0,
            'failure_count': 0,
            'job_state': 'Error',
            'success': False,
            'error': str(e)
        }

# Legacy function for backward compatibility
def successful_results(access_info, job_id):
    """Legacy function - now redirects to the new comprehensive processing."""
    print("⚠️ Warning: This function is deprecated. Use process_bulk_api_results() instead.")
    return get_successful_results(access_info, job_id)


import re
import requests
import pandas as pd
from urllib.parse import unquote_plus

def remove_field_from_query(query, field_to_remove):
    """
    Remove a specific field from a SOQL query string.
    
    Args:
        query: SOQL query string (e.g., "SELECT Id, Name, Field1 FROM Account")
        field_to_remove: Field name to remove from the query
    
    Returns:
        str: Modified query string with the field removed
    """
    print(f"🔍 DEBUG: Original query: '{query}'")
    print(f"🔍 DEBUG: Query type: {type(query)}")
    print(f"🔍 DEBUG: Query length: {len(query) if query else 0}")
    print(f"🔍 DEBUG: Field to remove: '{field_to_remove}'")
    
    # Check if query is None or empty
    if not query or not isinstance(query, str):
        print(f"❌ DEBUG: Query is None, empty, or not a string")
        return query or ""
    
    # Clean up the query string and handle URL encoding
    query_clean = query.strip()
    if not query_clean:
        print(f"❌ DEBUG: Query is empty after stripping whitespace")
        return query
    
    # Check if query is URL encoded (contains + instead of spaces)
    is_url_encoded = '+' in query_clean and not re.search(r'SELECT\s+.*\s+FROM', query_clean, re.IGNORECASE)
    print(f"🔍 DEBUG: Query appears to be URL encoded: {is_url_encoded}")
    
    if is_url_encoded:
        # URL decode the query by replacing + with spaces
        query_for_parsing = unquote_plus(query_clean)
        print(f"🔍 DEBUG: URL decoded query (first 200 chars): '{query_for_parsing[:200]}...'")
    else:
        query_for_parsing = query_clean
        print(f"🔍 DEBUG: Query for parsing (first 200 chars): '{query_for_parsing[:200]}...'")
    
    # Try multiple regex patterns to be more flexible
    patterns = [
        r'SELECT\s+(.*?)\s+FROM',  # Standard pattern
        r'SELECT\s+(.*?)\sFROM',   # Less strict whitespace
        r'SELECT\s+(.*?)FROM',     # Even less strict
        r'(?i)SELECT\s+(.*?)\s+FROM',  # Case insensitive inline
        r'SELECT[\s+]+(.*?)[\s+]+FROM',  # Handle + or spaces
    ]
    
    select_match = None
    for i, pattern in enumerate(patterns):
        select_match = re.search(pattern, query_for_parsing, re.IGNORECASE | re.DOTALL)
        if select_match:
            print(f"✅ DEBUG: Pattern {i+1} matched: {pattern}")
            break
        else:
            print(f"❌ DEBUG: Pattern {i+1} failed: {pattern}")
    
    if not select_match:
        print(f"❌ DEBUG: Could not parse SELECT clause from query")
        print(f"🔍 DEBUG: Query contains 'SELECT'? {'SELECT' in query_for_parsing.upper()}")
        print(f"🔍 DEBUG: Query contains 'FROM'? {'FROM' in query_for_parsing.upper()}")
        
        # Show character-by-character breakdown for very short queries
        if len(query_for_parsing) < 100:
            print(f"🔍 DEBUG: Character breakdown: {[c for c in query_for_parsing]}")
        
        return query
    
    select_clause = select_match.group(1)
    print(f"🔍 DEBUG: SELECT clause (first 200 chars): {select_clause[:200]}...")
    
    # Always clean the select clause properly - remove + signs and extra spaces
    select_clause_clean = select_clause.replace('+', ' ')
    select_clause_clean = re.sub(r'\s+', ' ', select_clause_clean.strip())
    
    # Split by commas and clean each field
    fields = [field.strip() for field in select_clause_clean.split(',') if field.strip()]
    print(f"🔍 DEBUG: Total fields found: {len(fields)}")
    print(f"🔍 DEBUG: First 10 fields: {fields[:10]}")
    
    # Find the field to remove (case-insensitive)
    field_found = False
    for field in fields:
        field_name = field.split(' AS ')[0].split(' as ')[0].strip()
        if field_name.upper() == field_to_remove.upper():
            field_found = True
            break
    
    print(f"🔍 DEBUG: Field '{field_to_remove}' found in query: {field_found}")
    
    # Remove the specified field (case-insensitive)
    # Also handle field aliases (e.g., "Field__c AS MyField")
    cleaned_fields = []
    field_removed = False
    
    for field in fields:
        # Extract the actual field name (before any alias)
        field_name = field.split(' AS ')[0].split(' as ')[0].strip()
        
        if field_name.upper() != field_to_remove.upper():
            cleaned_fields.append(field)
        else:
            field_removed = True
            print(f"✅ DEBUG: Removed field: {field}")
    
    print(f"🔍 DEBUG: Fields after removal: {len(cleaned_fields)} remaining")
    print(f"🔍 DEBUG: Original field count: {len(fields)}, Cleaned field count: {len(cleaned_fields)}")
    
    if not field_removed:
        print(f"❌ DEBUG: Field {field_to_remove} not found in fields list")
        return query
    
    if len(cleaned_fields) == 0:
        print(f"❌ DEBUG: Cannot remove all fields from SELECT clause")
        return query
    
    # Reconstruct the query
    if is_url_encoded:
        # Put back the + signs for URL encoded format
        cleaned_select = ', '.join(cleaned_fields).replace(' ', '+')
        # For URL encoded queries, we need to be more careful with the replacement
        cleaned_query = re.sub(
            r'(select\+).*?(\+from)', 
            rf'\1{cleaned_select}\2', 
            query, 
            flags=re.IGNORECASE
        )
    else:
        cleaned_select = ', '.join(cleaned_fields)
        cleaned_query = re.sub(
            r'(SELECT\s+).*?(\s+FROM)', 
            rf'\1{cleaned_select}\2', 
            query, 
            flags=re.IGNORECASE | re.DOTALL
        )
    
    print(f"🔍 DEBUG: Cleaned query (first 200 chars): {cleaned_query[:200]}...")
    return cleaned_query


def query_records(access_info, query, batch_size=1000, incremental=False):
    """
    Query Salesforce records and yield DataFrames for each batch.
    
    Args:
        access_info: Dictionary that contains Salesforce 'access_token' and 'instance_url'.
        query: Salesforce SOQL query string.
        batch_size: Number of records per batch (default: 1000).  Can be up to 2000.
    
    Yields:
        pandas.DataFrame: DataFrame containing a batch of records.
    
    Returns:
        None: If no records are found.
    """

    headers = {
        "Authorization": f"Bearer {access_info['access_token']}",
        "Content-Type": "application/json",
        "Sforce-Query-Options": f"batchSize={batch_size}"
    }

    url = f"{access_info['instance_url']}/services/data/v58.0/queryAll?q={query}"
    print("@@@ QUERY: ", query)
    
    results = requests.get(url, headers=headers)
    
    # Check for INVALID_FIELD errors and retry with cleaned query
    if results.status_code == 400:
        json_data = results.json()
        if isinstance(json_data, list) and len(json_data) > 0:
            error_info = json_data[0]
            if error_info.get('errorCode') == 'INVALID_FIELD':
                error_message = error_info.get('message', '')
                print(f"❌ INVALID_FIELD error: {error_message}")
                
                # Extract field name from error message
                field_match = re.search(r"No such column '([^']+)'", error_message)
                if field_match:
                    invalid_field = field_match.group(1)
                    print(f"🔍 Removing invalid field: {invalid_field}")
                    
                    # Remove the invalid field from the query
                    cleaned_query = remove_field_from_query(query, invalid_field)
                    if cleaned_query != query:
                        print(f"🔄 Retrying with cleaned query")
                        # Use 'yield from' to properly delegate the generator
                        yield from query_records(access_info, cleaned_query, batch_size, incremental)
                        return  # Exit early after successful retry
                    else:
                        print(f"❌ Could not remove field {invalid_field} from query")
    
    results.raise_for_status()  # Raise exception for HTTP errors
    json_data = results.json()

    if json_data['totalSize'] == 0:
        return None

    sobj_data = pd.json_normalize(json_data['records'])
    try:
        sobj_data.drop(columns=['attributes.type', 'attributes.url'], inplace=True)
    except KeyError:
        print("Attributes not found, moving on")
    
    for col in sobj_data.select_dtypes(include=['datetime64']).columns:
        sobj_data[col] = sobj_data[col].fillna(pd.Timestamp('1900-01-01'))

    for col in sobj_data.select_dtypes(include=['float64', 'int64']).columns:
        sobj_data[col] = sobj_data[col].fillna(0)

    for col in sobj_data.select_dtypes(include=['object']).columns:
        sobj_data[col] = sobj_data[col].fillna('')

    sobj_data.columns = sobj_data.columns.str.upper()

    yield sobj_data

    while json_data.get('nextRecordsUrl'):
        url = f"{access_info['instance_url']}{json_data['nextRecordsUrl']}"
        results = requests.get(url, headers=headers)
        results.raise_for_status()
        json_data = results.json()

        sobj_data = pd.json_normalize(json_data['records'])
        sobj_data.columns = sobj_data.columns.str.upper()
        try:
            sobj_data.drop(columns=['ATTRIBUTES.TYPE', 'ATTRIBUTES.URL'], inplace=True)
        except KeyError:
            print("Attributes not found, moving on")
        
        for col in sobj_data.select_dtypes(include=['datetime64']).columns:
            sobj_data[col] = sobj_data[col].fillna(pd.Timestamp('1900-01-01')) 
        for col in sobj_data.select_dtypes(include=['float64', 'int64']).columns:
            sobj_data[col] = sobj_data[col].fillna(0)
        for col in sobj_data.select_dtypes(include=['object']).columns:
            sobj_data[col] = sobj_data[col].fillna('')

        yield sobj_data


import pandas as pd
import numpy as np
import tempfile
import re


def salesforce_field_type(field_type):
	if field_type['type'] == 'id':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'boolean':
		return 'boolean'
	elif field_type['type'] == 'reference':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'string':
		return 'string' #'string({})'.format(field_type['length']) -- modified this because mixed types calls strings that look like numbers to overflow, ie '20' becomes 20.0 even when it's converted back to a string
	elif field_type['type'] == 'email':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'picklist':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'textarea':
		return 'string'
	elif field_type['type'] == 'double':
		if field_type['precision'] > 0:
			precision = field_type['precision']
		elif field_type['digits'] > 0:
			precision = field_type['precision']
		scale = field_type['scale']
		return 'NUMBER({},{})'.format(precision, scale)
	elif field_type['type'] == 'phone':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'datetime':
		return 'timestamp_ntz' #'NUMBER(38,0)' #
	elif field_type['type'] == 'date':
		return 'date' #'NUMBER(38,0)' #
	elif field_type['type'] == 'address':
		return 'string' #({})'.format(field_type['length'])
	elif field_type['type'] == 'url':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'currency':
		return 'number({},{})'.format(field_type['precision'], field_type['scale'])
	elif field_type['type'] == 'int':
		if field_type['precision'] > 0:
			precision = field_type['precision']
		elif field_type['digits'] > 0:
			precision = field_type['digits']
		return 'number({},{})'.format(precision, field_type['scale'])
	elif field_type['type'] == 'multipicklist':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'percent':
		return 'number({},{})'.format(field_type['precision'], field_type['scale'])
	elif field_type['type'] == 'combobox':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'encryptedstring':
		return 'string({})'.format(field_type['length'])
	elif field_type['type'] == 'base64':
		return 'string'
	elif field_type['type'] == 'datacategorygroupreference':
		return 'string(80)'	
	elif field_type['type'] == 'anyType':
		return 'string'
	elif field_type['type'] == 'byte':
		return 'string(1)'	
	elif field_type['type'] == 'calc':
		return 'string(255)'
	# Removed duplicate int case
	elif field_type['type'] == 'junctionidlist':
		return 'string(18)'
	elif field_type['type'] == 'long':
		return 'number(32)'
	elif field_type['type'] == 'time':
		return 'string(24)'
	else:
		print("KACK {}".format(field_type['type']))
		exit(0)
	
def df_field_type(field_type):
	if field_type['type'] == 'id':
		return 'object'
	elif field_type['type'] == 'boolean':
		return 'bool'
	elif field_type['type'] == 'reference':
		return 'object'
	elif field_type['type'] == 'string':
		return 'object'
	elif field_type['type'] == 'email':
		return 'object'
	elif field_type['type'] == 'picklist':
		return 'object'
	elif field_type['type'] == 'textarea':
		return 'object'
	elif field_type['type'] == 'double':
		return 'float64'
	elif field_type['type'] == 'phone':
		return 'object'
	elif field_type['type'] == 'datetime':
		return 'datetime64'
	elif field_type['type'] == 'date':
		return 'date'  # Keep date fields as date, not convert to datetime64
	elif field_type['type'] == 'address':
		return 'object'
	elif field_type['type'] == 'url':
		return 'object'
	elif field_type['type'] == 'currency':
		return 'float64'
	elif field_type['type'] == 'int':
		return 'int64'

def convert_field_types(df, df_fieldsets, table_fields):

	for col, dtype in df_fieldsets.items():

		if col.upper() not in table_fields:
			df.drop(columns=[col], inplace=True)
			continue 
		elif dtype == 'date':
			df[col] == pd.to_datetime(df[col],errors='coerce').dt.date
		elif dtype == 'int64':
			df[col] = pd.to_numeric(df[col], errors='coerce').astype('Int64')
		elif dtype == 'object':
			df[col] = df[col].astype(str)
		elif dtype == 'float64':
			df[col] = pd.to_numeric(df[col], errors='coerce').astype('float64')
		elif dtype == 'bool':
			df[col] = pd.to_numeric(df[col], errors='coerce').astype('bool')
		elif dtype == 'datetime64':
			df[col] = pd.to_datetime(df[col], errors='coerce')
			#df[col] = pd.to_datetime(df[col],errors='coerce').dt.datetime64
	df = df.replace(np.nan, None)
	return df.rename(columns={col: col.upper() for col in df.columns})

def cache_data(data):
	with tempfile.NamedTemporaryFile(delete=False) as temp_file:
		temp_file.write(data)
		temp_file_path = temp_file.name
	#print("  {}".format(temp_file_path))
	return temp_file_path


def format_sync_file(df, df_fields):
	"""
	Format DataFrame for Salesforce sync operations with robust type handling.
	
	This function now includes improved DATE field handling that prevents the
	'str object is not callable' error by ensuring proper type conversion
	and fallback handling for problematic data.
	"""
	# First, convert all column names to uppercase to match Salesforce API response
	df.columns = df.columns.str.upper()
	
	# Helper function to safely convert to string
	def safe_to_string(series):
		if series.dtype != 'object':
			return series.astype(str)
		return series
	
	# Pre-process all columns to ensure they're in a clean state
	print("🔍 Pre-processing DataFrame columns for type safety...")
	for col in df.columns:
		try:
			# Handle any problematic data types before main processing
			if df[col].dtype == 'object':
				# Clean up object columns that might have mixed types
				df[col] = df[col].replace({pd.NA: None, pd.NaT: None, 'nan': None, 'None': None, '<NA>': None})
				df[col] = safe_to_string(df[col])
				#print(f"✅ Pre-processed column {col} to clean string")
		except Exception as e:
			print(f"⚠️ Warning: Error pre-processing column {col}: {e}")
			# Fallback: ensure it's a string
			try:
				df[col] = df[col].astype(str)
			except:
				# Last resort: fill with None
				df[col] = None
				print(f"⚠️ Column {col} filled with None due to conversion error")
	
	print("✅ DataFrame pre-processing completed")
	
	# Debug: Show field type mappings
	print("🔍 Field type mappings:")
	for col, dtype in df_fields.items():
		print(f"  {col}: {dtype}")
	print()
	
	# Debug: Check if EFFECTIVEDATE is in the mappings
	if 'EFFECTIVEDATE' in df_fields:
		print(f"✅ EFFECTIVEDATE found in df_fields with type: {df_fields['EFFECTIVEDATE']}")
	else:
		print("❌ EFFECTIVEDATE NOT found in df_fields")
		# Check for case variations
		effective_date_variations = [col for col in df_fields.keys() if col.upper() == 'EFFECTIVEDATE']
		if effective_date_variations:
			print(f"✅ Found EFFECTIVEDATE variations: {effective_date_variations}")
			for var in effective_date_variations:
				print(f"  {var}: {df_fields[var]}")
		else:
			print("❌ No EFFECTIVEDATE variations found either")
		print(f"Available fields: {list(df_fields.keys())}")
	
	# Debug: Show all datetime fields
	datetime_fields = [col for col, dtype in df_fields.items() if dtype in ['datetime64', 'datetime64[ns]']]
	if datetime_fields:
		print(f"🔍 Datetime fields found: {datetime_fields}")
	else:
		print("⚠️ No datetime fields found in df_fields")
	print()
	
	# Now process each field according to its intended type
	for col, dtype in df_fields.items():
		# Convert field name to uppercase to match DataFrame columns
		col_upper = col.upper()
		
		# Debug: Show what's happening with EFFECTIVEDATE specifically
		if col_upper == 'EFFECTIVEDATE':
			print(f"🔍 Processing EFFECTIVEDATE field:")
			print(f"  Original col name: {col}")
			print(f"  Mapped dtype: {dtype}")
			print(f"  Uppercase col: {col_upper}")
			print(f"  In DataFrame columns: {col_upper in df.columns}")
			if col_upper in df.columns:
				print(f"  DataFrame dtype: {df[col_upper].dtype}")
				print(f"  Sample values: {df[col_upper].head(3).tolist()}")
		
		# Case-insensitive field matching
		# First try exact match, then try uppercase match
		field_found = False
		if col in df.columns:
			col_upper = col  # Use original case
			field_found = True
		elif col_upper in df.columns:
			field_found = True
		else:
			# Try to find a case-insensitive match
			for df_col in df.columns:
				if df_col.upper() == col_upper:
					col_upper = df_col  # Use the actual DataFrame column name
					field_found = True
					break
		
		try:
			if field_found:
				# Additional safety check: ensure column is in a good state before processing
				if df[col_upper].dtype == 'object' and df[col_upper].isna().all():
					#print(f"⚠️ Warning: Column {col_upper} is all NaN, skipping type conversion")
					continue
				# CRITICAL: Force fields to their intended types BEFORE any data analysis
				# This ensures write_pandas creates the correct table schema
				if dtype == 'datetime64' or dtype == 'datetime64[ns]':
					# Salesforce datetime fields (like CreatedDate, LastViewedDate, LastActivityDate) MUST be datetime
					print(f"🔍 Processing datetime field: {col_upper}")
					print(f"  Original dtype: {df[col_upper].dtype}")
					print(f"  Sample values: {df[col_upper].head(3).tolist()}")
					
					# First, handle any None/NaN values that might cause conversion issues
					df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None, 'nan': None, 'None': None, '<NA>': None})
					
					# Convert to datetime with more robust error handling
					try:
						df[col_upper] = pd.to_datetime(df[col_upper], errors='coerce', utc=True)
						
						# Fill NaT values with a default datetime
						df[col_upper] = df[col_upper].fillna(pd.Timestamp('1900-01-01 00:00:00'))
						
						# Check if we have timezone-aware datetimes and convert to timezone-naive
						if hasattr(df[col_upper], 'dt') and hasattr(df[col_upper].dt, 'tz'):
							tz_info = df[col_upper].dt.tz
							if tz_info is not None:
								df[col_upper] = df[col_upper].dt.tz_localize(None)
						
						# Verify we have a proper datetime column
						if pd.api.types.is_datetime64_any_dtype(df[col_upper]):
							print(f"✅ Successfully converted {col_upper} to datetime64")
						else:
							print(f"⚠️ Warning: {col_upper} is not datetime64 after conversion, dtype: {df[col_upper].dtype}")
							# Try one more time with string conversion first
							df[col_upper] = pd.to_datetime(df[col_upper].astype(str), errors='coerce', utc=True)
							df[col_upper] = df[col_upper].fillna(pd.Timestamp('1900-01-01 00:00:00'))
							
					except Exception as e:
						print(f"⚠️ Warning: Error converting {col_upper} to datetime: {e}")
						# Fallback: convert to string to preserve data
						df[col_upper] = safe_to_string(df[col_upper])
						df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
					# # Salesforce datetime/date fields (like CreatedDate, LastViewedDate, LastActivityDate) MUST be datetime
					# print(f"🔍 Processing datetime field: {col_upper}")
					# print(f"  Original dtype: {df[col_upper].dtype}")
					# print(f"  Sample values: {df[col_upper].head(3).tolist()}")
					
					# # First, handle any None/NaN values that might cause conversion issues
					# df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None, 'nan': None, 'None': None, '<NA>': None})
					
					# # Convert to datetime with more robust error handling
					# try:
					# 	df[col_upper] = pd.to_datetime(df[col_upper], errors='coerce', utc=True)
						
					# 	# Fill NaT values with a default datetime
					# 	df[col_upper] = df[col_upper].fillna(pd.Timestamp('1900-01-01 00:00:00'))
						
					# 	# Check if we have timezone-aware datetimes and convert to timezone-naive
					# 	if hasattr(df[col_upper], 'dt') and hasattr(df[col_upper].dt, 'tz'):
					# 		tz_info = df[col_upper].dt.tz
					# 		if tz_info is not None:
					# 			df[col_upper] = df[col_upper].dt.tz_localize(None)
						
					# 	# Verify we have a proper datetime column
					# 	if pd.api.types.is_datetime64_any_dtype(df[col_upper]):
					# 		print(f"✅ Successfully converted {col_upper} to datetime64")
					# 	else:
					# 		print(f"⚠️ Warning: {col_upper} is not datetime64 after conversion, dtype: {df[col_upper].dtype}")
					# 		# Try one more time with string conversion first
					# 		df[col_upper] = pd.to_datetime(df[col_upper].astype(str), errors='coerce', utc=True)
					# 		df[col_upper] = df[col_upper].fillna(pd.Timestamp('1900-01-01 00:00:00'))
							
					# except Exception as e:
					# 	print(f"⚠️ Warning: Error converting {col_upper} to datetime: {e}")
					# 	# Fallback: convert to string to preserve data
					# 	df[col_upper] = safe_to_string(df[col_upper])
					# 	df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
						
				elif dtype == 'date':
					# Handle date fields - just fill None values, don't change the type
					df[col_upper] = df[col_upper].fillna(pd.Timestamp('1900-01-01').date())

				elif dtype == 'object':
					# Salesforce string fields (including PO_Number__c) MUST be strings
					# Convert to string immediately, regardless of content
					df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None})
					df[col_upper] = safe_to_string(df[col_upper])
					df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
					
				elif dtype == 'int64':
					# Check if ANY value is non-numeric - if so, convert entire column to string
					has_non_numeric = False
					for value in df[col_upper].dropna():
						if isinstance(value, str) and not value.replace('-', '').replace('.', '').isdigit():
							has_non_numeric = True
							break
					
					if has_non_numeric:
						# Convert entire column to string - no mixed types allowed in Snowflake
						df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None})
						df[col_upper] = safe_to_string(df[col_upper])
						df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
					else:
						# All values are numeric, safe to convert
						try:
							df[col_upper] = pd.to_numeric(df[col_upper], errors='coerce').astype('Int64')
						except Exception as e:
							print(f"⚠️ Warning: Could not convert {col_upper} to int64, treating as string: {e}")
							df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None})
							df[col_upper] = safe_to_string(df[col_upper])
							df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
							
				elif dtype == 'float64':
					# For float fields (like latitude/longitude), handle empty strings properly
					# First, replace empty strings with None (which pandas can handle)
					df[col_upper] = df[col_upper].replace({'': None, 'nan': None, 'None': None, '<NA>': None})
					
					# Now convert to float64 - this will convert None to NaN, which is fine
					try:
						df[col_upper] = pd.to_numeric(df[col_upper], errors='coerce').astype('float64')
					except Exception as e:
						print(f"❌ Failed to convert {col_upper} to float64: {e}")
						# Fallback: convert to string but preserve None values
						df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None})
						df[col_upper] = safe_to_string(df[col_upper])
						df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
							
				elif dtype == 'bool':
					# Check for non-boolean values
					has_non_bool = False
					for value in df[col_upper].dropna():
						if isinstance(value, str) and value.lower() not in ['true', 'false', '1', '0', 'yes', 'no']:
							has_non_bool = True
							break
					
					if has_non_bool:
						# Convert entire column to string
						print(f"⚠️ Column {col_upper} contains non-boolean values, converting entire column to string")
						df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None})
						df[col_upper] = safe_to_string(df[col_upper])
						df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
					else:
						# All values are boolean-like, safe to convert
						try:
							df[col_upper] = pd.to_numeric(df[col_upper], errors='coerce').astype('bool')
						except Exception as e:
							print(f"⚠️ Warning: Could not convert {col_upper} to bool, treating as string: {e}")
							df[col_upper] = df[col_upper].replace({pd.NA: None, pd.NaT: None})
							df[col_upper] = safe_to_string(df[col_upper])
							df[col_upper] = df[col_upper].replace({'nan': None, 'None': None, '<NA>': None})
							

			else:
				print(f"field not found '{col_upper}' in DataFrame columns: {list(df.columns)}")
		except Exception as e:
			print(f"field not found '{col_upper}': {e}")
	
	# Final safety check: ensure all columns are in a safe state for Snowpark conversion
	print("🔍 Final safety check: ensuring all columns are Snowpark-compatible...")
	for col in df.columns:
		try:
			# Ensure the column is in a safe state
			if df[col].dtype == 'object':
				# For object columns, ensure they're clean strings
				df[col] = df[col].replace({pd.NA: None, pd.NaT: None, 'nan': None, 'None': None, '<NA>': None})
				# Convert to string if not already
				if df[col].dtype != 'object':
					df[col] = df[col].astype(str)
			elif df[col].dtype in ['int64', 'float64']:
				# For numeric columns, handle NaN values
				if df[col].isna().any():
					df[col] = df[col].fillna(0)
		except Exception as e:
			print(f"⚠️ Warning: Error in final safety check for column {col}: {e}")
			# Last resort: convert to string
			try:
				df[col] = df[col].astype(str)
			except:
				df[col] = None
	
	print("✅ Final safety check completed")
	return df

def map_table_field_types(schema_df, dataframe, df_cols):
    df_cols = [col.upper() for col in df_cols]
    
    # Initialize Pandas_Type column
    schema_df['Pandas_Type'] = None
    
    # Track columns to keep
    cols_to_keep = [col for col in dataframe.columns if col.upper() in df_cols]
    
    # Drop columns not in df_cols
    cols_to_drop = [col for col in dataframe.columns if col.upper() not in df_cols]
    if cols_to_drop:
        try:
            dataframe = dataframe.drop(cols_to_drop)
            print(f"Dropped columns not in df_cols: {cols_to_drop}")
        except Exception as e:
            print(f"Error dropping columns: {e}")
    
    # Map types
    for index, row in schema_df.iterrows():
        col = row['Column_Name'].upper()
        dtype = row['Data_Type']
        # Skip columns not in df_cols
        try:
            if col not in df_cols:
                dataframe = dataframe.drop(col, axis=1, inplace=True)
        except Exception as e:
            pass

        # Map Salesforce/Snowflake types to Pandas types
        if dtype.startswith('StringType') or dtype.startswith('VARCHAR'):
            schema_df.at[index, 'Pandas_Type'] = 'object'
        elif dtype.startswith('BooleanType') or dtype.startswith('BOOLEAN'):
            schema_df.at[index, 'Pandas_Type'] = 'bool'
        elif dtype.startswith('DecimalType') or dtype.startswith('NUMBER'):
            schema_df.at[index, 'Pandas_Type'] = 'float64'
        elif dtype.startswith('LongType') or dtype.startswith('INTEGER'):
            schema_df.at[index, 'Pandas_Type'] = 'int64'
        elif dtype.startswith('TimestampType') or dtype.startswith('TIMESTAMP'):
            schema_df.at[index, 'Pandas_Type'] = 'datetime64[ns]'
        elif dtype.startswith('DateType') or dtype.startswith('DATE'):
            schema_df.at[index, 'Pandas_Type'] = 'date'
        else:
            schema_df.at[index, 'Pandas_Type'] = 'unknown'
    
    return schema_df, dataframe


from .soql_query import build_soql  # If this is how you want to import it
from .csv import json_to_csv
from .data_writer import (
    write_dataframe_to_table, 
    write_batch_to_temp_table, 
    write_batch_to_main_table,
    validate_dataframe_types,
    standardize_dataframe_types,
    write_dataframe_with_type_handling
)
from .table_creator import (
    create_salesforce_table,
    ensure_table_exists_for_dataframe
)



import requests
import json
from lht.util import csv
from lht.sflake import query as q
from lht.salesforce import ingest_bapi20 as ingest
import time

def upsert(session, access_info, sobject, query, field, batch_size=25000):
    """
    Upsert records to Salesforce using data from a SQL query executed against Snowflake.
    Processes records in batches to handle large datasets efficiently.
    
    Args:
        session: Snowflake session object
        access_info: Salesforce access credentials dictionary
        sobject: Salesforce object name (e.g., 'Account', 'Contact')
        query: SQL query string to execute against Snowflake
        field: External ID field name for upsert operation
        batch_size: Number of records to process per batch (default: 25000)
    """
    print("\n" + "="*100)
    print("🚀 STARTING SALESFORCE UPSERT WITH BATCH PROCESSING")
    print("="*100)
    print(f"📋 Parameters:")
    print(f"   - SObject: {sobject}")
    print(f"   - External ID Field: {field}")
    print(f"   - Batch Size: {batch_size:,} records per batch")
    print(f"   - SQL Query: {query[:100]}{'...' if len(query) > 100 else ''}")
    print(f"   - Instance URL: {access_info.get('instance_url', 'Not available')}")
    print("="*100)
    
    try:
        access_token = access_info['access_token']
        
        print("🔍 STEP 1: Getting total record count...")
        # First, get the total count of records to determine number of batches
        count_query = f"SELECT COUNT(*) as total_count FROM ({query})"
        count_result = session.sql(count_query).collect()
        total_records = count_result[0][0] if count_result else 0
        
        print(f"📊 Total records to process: {total_records:,}")
        
        if total_records == 0:
            print("⚠️ No records found to process")
            return None
        
        # Calculate number of batches needed
        num_batches = (total_records + batch_size - 1) // batch_size
        print(f"📦 Will process {total_records:,} records in {num_batches} batch(es) of up to {batch_size:,} records each")
        
        bulk_api_url = access_info['instance_url']+ f"/services/data/v62.0/jobs/ingest"
        print(f"🔗 Bulk API URL: {bulk_api_url}")

        # Job data template
        job_data_template = {
            "object": f"{sobject}",  # Specify the Salesforce object
            "operation": "upsert",  # Use upsert operation
            "externalIdFieldName": f"{field}",  # Field to use for upsert
            "lineEnding" : "CRLF"
        }

        headers = {'Authorization': f'Bearer {access_token}', 'Content-Type': 'application/json'}
        
        all_job_info = []
        successful_batches = 0
        failed_batches = 0
        total_processed = 0

        # Process records in batches - query each batch separately
        for batch_num in range(num_batches):
            offset = batch_num * batch_size
            limit = batch_size
            
            print(f"\n🔍 STEP 2.{batch_num + 1}: Processing batch {batch_num + 1}/{num_batches}")
            print(f"📊 Batch range: records {offset + 1:,} to {min(offset + batch_size, total_records):,}")
            
            try:
                # Query only the records for this batch using LIMIT and OFFSET
                batch_query = f"{query} LIMIT {limit} OFFSET {offset}"
                print(f"🔍 Executing batch query: {batch_query[:100]}...")
                
                results = session.sql(batch_query).collect()
                
                # Convert results to the expected format
                batch_records = []
                for result in results:
                    record = {}
                    for key, value in result.asDict().items():
                        if value is None:
                            record[key] = ''
                        else:
                            record[key] = value
                    batch_records.append(record)
                
                actual_batch_size = len(batch_records)
                total_processed += actual_batch_size
                print(f"📊 Retrieved {actual_batch_size:,} records for this batch (Total processed: {total_processed:,}/{total_records:,})")
                
                if actual_batch_size == 0:
                    print("⚠️ No records in this batch, skipping...")
                    continue
                
                print("🔍 Converting batch records to CSV format...")
                batch_data = csv.json_to_csv(batch_records)
                print(f"📄 Batch CSV data length: {len(batch_data):,} characters")

                # Create a new job for this batch
                print("🔍 Creating Salesforce Bulk API job for this batch...")
                response = requests.post(bulk_api_url, headers=headers, data=json.dumps(job_data_template))
                
                if response.status_code != 200:
                    print(f"❌ Job creation failed with status {response.status_code}")
                    print(f"❌ Response: {response.text}")
                    response.raise_for_status()
                    
                job_info = response.json()
                print(f"✅ Job created successfully: {job_info}")
                job_id = job_info['id']
                print(f"🆔 Job ID: {job_id}")

                #########################################################
                ###  SEND BATCH FILE
                #########################################################
                print("🔍 Sending batch CSV data to Salesforce...")
                ingest.send_file(access_info, job_id, batch_data)
                print("✅ Batch file sent successfully")
                
                #########################################################
                ###  CLOSE JOB
                #########################################################
                print("🔍 Closing job to start processing...")
                close_results = ingest.job_close(access_info, job_id)
                print(f"✅ Job closed: {close_results}")

                #########################################################
                ###  CHECK STATUS
                #########################################################
                print("🔍 Monitoring batch job status...")
                status_check_count = 0
                batch_success = False
                
                while True:
                    status_check_count += 1
                    close_results = ingest.job_status(access_info, job_id)
                    print(f"📊 Status check #{status_check_count} - ID: {close_results['id']}, Status: {close_results['state']}")
                    
                    if close_results['state'] == 'JobComplete':
                        print("✅ Batch job completed successfully!")
                        batch_success = True
                        successful_batches += 1
                        break
                    elif close_results['state'] in ['Failed', 'Aborted']:
                        print(f"❌ Batch job failed with status: {close_results['state']}")
                        print(f"❌ Full job details: {close_results}")
                        failed_batches += 1
                        break
                    
                    print("⏳ Waiting 10 seconds before next status check...")
                    time.sleep(10)
                
                # Store job info for this batch
                batch_job_info = {
                    'batch_number': batch_num + 1,
                    'job_id': job_id,
                    'records_processed': actual_batch_size,
                    'success': batch_success,
                    'job_info': job_info
                }
                all_job_info.append(batch_job_info)
                
                print(f"✅ Batch {batch_num + 1} processing completed")
                
                # Clear batch_records from memory to free up space
                del batch_records
                del batch_data
                
            except Exception as batch_error:
                print(f"❌ Error processing batch {batch_num + 1}: {batch_error}")
                failed_batches += 1
                batch_job_info = {
                    'batch_number': batch_num + 1,
                    'job_id': None,
                    'records_processed': 0,
                    'success': False,
                    'error': str(batch_error)
                }
                all_job_info.append(batch_job_info)
                continue

        print("\n" + "="*100)
        print("✅ BATCH UPSERT PROCESSING COMPLETED")
        print("="*100)
        print(f"📊 Summary:")
        print(f"   - Total records: {total_records:,}")
        print(f"   - Records processed: {total_processed:,}")
        print(f"   - Total batches: {num_batches}")
        print(f"   - Successful batches: {successful_batches}")
        print(f"   - Failed batches: {failed_batches}")
        print(f"   - Success rate: {(successful_batches/num_batches)*100:.1f}%" if num_batches > 0 else "   - Success rate: N/A")
        print("="*100)
        
        return {
            'total_records': total_records,
            'records_processed': total_processed,
            'total_batches': num_batches,
            'successful_batches': successful_batches,
            'failed_batches': failed_batches,
            'batch_results': all_job_info
        }
        
    except Exception as e:
        print("\n" + "="*100)
        print("❌ UPSERT FAILED")
        print("="*100)
        print(f"Error: {e}")
        print(f"Error type: {type(e).__name__}")
        import traceback
        print(f"Full traceback:")
        print(traceback.format_exc())
        print("="*100)
        raise

def update(session, access_info, sobject, query):
    access_token = access_info['access_token']

    #records = q.get_records(session, query)
    results = session.sql(query).collect()
    # Convert results to the expected format
    records = []
    for result in results:
        record = {}
        for key, value in result.asDict().items():
            if value is None:
                record[key] = ''
            else:
                record[key] = value
        records.append(record)

    data = csv.json_to_csv(records)

    bulk_api_url = access_info['instance_url']+ f"/services/data/v62.0/jobs/ingest"

    # Create a new job
    job_data = {
        "object": f"{sobject}",  # Specify the Salesforce object
        "operation": "update",  # Use upsert operation
        "lineEnding" : "CRLF"
    }

    headers = {'Authorization': f'Bearer {access_token}', 'Content-Type': 'application/json'}

    # Create the job
    print("creating job")
    response = requests.post(bulk_api_url, headers=headers, data=json.dumps(job_data))
    job_info = response.json()
    #log_retl.job(session, job_info)

    job_id = job_info['id']

    #########################################################
    ###  SEND BATCH FILE
    #########################################################
    #def add_batch(instance_url, access_token, job_id, data):
    print("sending file")
    ingest.send_file(access_info, job_id, data)
    
    #########################################################
    ###  CLOSE JOB
    #########################################################
    print("closing job")
    close_results = ingest.job_close(access_info, job_id)
    print(close_results)


    #########################################################
    ###  CHECK STATUS
    #########################################################    
    while True:
        close_results = ingest.job_status(access_info, job_id)
        print("\nID: {}".format(close_results['id']))
        print("\nStatus: {}".format(close_results['state']))
        if close_results['state'] == 'JobComplete':
            break
        time.sleep(10)

    return job_info

def insert(session, access_info, sobject, query):
    access_token = access_info['access_token']

    records = q.get_records(session, query)
    data = csv.json_to_csv(records)

    bulk_api_url = access_info['instance_url']+ f"/services/data/v62.0/jobs/ingest"

    # Create a new job
    job_data = {
        "object": f"{sobject}",  
        "contentType" : "CSV",
        "operation": "insert",  
        "lineEnding" : "CRLF"
    }

    headers = {'Authorization': f'Bearer {access_token}', 'Content-Type': 'application/json'}

    # Create the job
    print("creating job")
    response = requests.post(bulk_api_url, headers=headers, data=json.dumps(job_data))
    job_info = response.json()
    #log_retl.job(session, job_info)

    job_id = job_info['id']

    #########################################################
    ###  SEND BATCH FILE
    #########################################################
    #def add_batch(instance_url, access_token, job_id, data):
    print("sending file")
    ingest.send_file(access_info, job_id, data)
    
    #########################################################
    ###  CLOSE JOB
    #########################################################
    print("closing job")
    close_results = ingest.job_close(access_info, job_id)
    print(close_results)


    #########################################################
    ###  CHECK STATUS
    #########################################################    
    while True:
        close_results = ingest.job_status(access_info, job_id)
        print("\nID: {}".format(close_results['id']))
        print("\nStatus: {}".format(close_results['state']))
        if close_results['state'] == 'JobComplete':
            break
        time.sleep(10)

    return job_info

def delete(session, access_info, sobject, query, field):

    access_token = access_info['access_token']

    results = session.sql(query).collect()
    # Convert results to the expected format
    records = []
    for result in results:
        record = {}
        for key, value in result.asDict().items():
            if value is None:
                record[key] = ''
            else:
                record[key] = value
        records.append(record)
    data = csv.json_to_csv(records)

    bulk_api_url = access_info['instance_url']+ f"/services/data/v62.0/jobs/ingest"

    # Create a new job
    job_data = {
        "object": f"{sobject}",  
        "contentType" : "CSV",
        "operation": "delete", 
        "lineEnding" : "CRLF"
    }

    headers = {'Authorization': f'Bearer {access_token}', 'Content-Type': 'application/json'}

    # Create the job
    print("creating job")
    response = requests.post(bulk_api_url, headers=headers, data=json.dumps(job_data))
    job_info = response.json()
    print("@@@ JOB: {}".format(job_info))
    #log_retl.job(session, job_info)

    job_id = job_info['id']

    #########################################################
    ###  SEND BATCH FILE
    #########################################################
    #def add_batch(instance_url, access_token, job_id, data):
    print("sending file")
    ingest.send_file(access_info, job_id, data)
    
    #########################################################
    ###  CLOSE JOB
    #########################################################
    print("closing job")
    close_results = ingest.job_close(access_info, job_id)
    print(close_results)


    #########################################################
    ###  CHECK STATUS
    #########################################################    
    while True:
        close_results = ingest.job_status(access_info, job_id)
        print("\nID: {}".format(close_results['id']))
        print("\nStatus: {}".format(close_results['state']))
        if close_results['state'] == 'JobComplete':
            break
        time.sleep(10)
    
    return job_info